//----------------------------------------------------------------------------
// NAME: Fog
// VERSION: 1.0
// DESCRIPTION: Add range fog. Adds itself as a layer in Layer Manager (key: L). Right click on layer for settings. 
// DEVELOPER: Patrick Murris
// WEBSITE: http://www.alpix.com/worldwin
//----------------------------------------------------------------------------
// Based on Bjorn Reppen 'Atmosphere' and 'RangeFog' plugin
// 1.1 June 19, 2005	Settings dialog, intensity select, read/save settings
// 1.0 June 15, 2005	First version
//----------------------------------------------------------------------------

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Windows.Forms;
using System.Drawing;

using WorldWind.Renderable;
using WorldWind.Camera;
using WorldWind;
using System.IO;
using System;

namespace Murris.Plugins
{
	/// <summary>
	/// The plugin (main class)
	/// </summary>
	public class Fog : WorldWind.PluginEngine.Plugin 
	{
		/// <summary>
		/// Name displayed in layer manager
		/// </summary>
		public static string LayerName = "Fog";

		/// <summary>
		/// Plugin entry point - All plugins must implement this function
		/// </summary>
		public override void Load() 
		{
			FogLayer layer = new FogLayer(LayerName, PluginDirectory, ParentApplication.WorldWindow);
			ParentApplication.WorldWindow.CurrentWorld.RenderableObjects.ChildObjects.Insert(0,layer);
		}

		/// <summary>
		/// Unloads our plugin
		/// </summary>
		public override void Unload() 
		{
			ParentApplication.WorldWindow.CurrentWorld.RenderableObjects.Remove(LayerName);
		}

	}

	/// <summary>
	/// Fog layer
	/// </summary>
	public class FogLayer : RenderableObject
	{
		static string version = "1.1";
		string settingsFileName = "fog.ini";
		string pluginPath;
		public World world;
		public DrawArgs drawArgs;
		Form pDialog;

		// Fog defaults
		float nearFactor = 2.0f;
		float farFactor = 15.0f;
		Color fogColor = Color.FromArgb(208, 208, 208);
		
		/// <summary>
		/// Constructor
		/// </summary>
		public FogLayer(string LayerName, string pluginPath, WorldWindow worldWindow) : base(LayerName)
		{
			this.pluginPath = pluginPath;
			this.world = worldWindow.CurrentWorld;
			this.drawArgs = worldWindow.DrawArgs;
			//this.RenderPriority = (RenderPriority)-1; // rendered before all others
			this.RenderPriority = RenderPriority.SurfaceImages;
			ReadSettings();
		}
		
		/// <summary>
		/// Read saved settings from ini file
		/// </summary>
		public void ReadSettings()
		{
			string line = "";
			try 
			{
				TextReader tr = File.OpenText(Path.Combine(pluginPath, settingsFileName));
				line = tr.ReadLine();
				tr.Close();
			}
			catch(Exception caught) {}
			if(line != "")
			{
				string[] settingsList = line.Split(';');
				string saveVersion = settingsList[1];	// version when settings where saved
				if(settingsList[1] != null) nearFactor = float.Parse(settingsList[1]);
				if(settingsList[2] != null) farFactor = float.Parse(settingsList[2]);
			}
		}

		/// <summary>
		/// Save settings in ini file
		/// </summary>
		public void SaveSettings()
		{
			string line = version + ";" + nearFactor.ToString() + ";" + farFactor.ToString();
			try
			{
				StreamWriter sw = new StreamWriter(Path.Combine(pluginPath, settingsFileName));
				sw.Write(line);
				sw.Close();
			}
			catch(Exception caught) {}
		}

		#region RenderableObject

		/// <summary>
		/// This is where we do our rendering 
		/// Called from UI thread = UI code safe in this function
		/// </summary>
		public override void Render(DrawArgs drawArgs)
		{
			if(!isInitialized)
				return;

			// Camera & Device shortcuts ;)
			CameraBase camera = drawArgs.WorldCamera;
			Device device = drawArgs.device;

			// Set fog params
			float a = (float)camera.Altitude;
			device.RenderState.FogEnable = true;
			device.RenderState.FogColor = fogColor;
			device.RenderState.FogTableMode = FogMode.Linear;
			device.RenderState.FogStart = a * nearFactor;
			device.RenderState.FogEnd = a * farFactor;
		}

		/// <summary>
		/// RenderableObject abstract member (needed) 
		/// OBS: Worker thread (don't update UI directly from this thread)
		/// </summary>
		public override void Initialize(DrawArgs drawArgs)
		{
			isInitialized = true;
		}

		/// <summary>
		/// RenderableObject abstract member (needed)
		/// OBS: Worker thread (don't update UI directly from this thread)
		/// </summary>
		public override void Update(DrawArgs drawArgs)
		{
			if(!isInitialized)
				Initialize(drawArgs);
		}

		/// <summary>
		/// RenderableObject abstract member (needed)
		/// OBS: Worker thread (don't update UI directly from this thread)
		/// </summary>
		public override void Dispose()
		{
			isInitialized = false;
		}

		/// <summary>
		/// Gets called when user left clicks.
		/// RenderableObject abstract member (needed)
		/// Called from UI thread = UI code safe in this function
		/// </summary>
		public override bool PerformSelectionAction(DrawArgs drawArgs)
		{
			return false;
		}

 		/// <summary>
 		/// Fills the context menu with menu items specific to the layer.
 		/// </summary>
 		public override void BuildContextMenu( ContextMenu menu )
 		{
  			menu.MenuItems.Add("Properties", new System.EventHandler(OnPropertiesClick));
 		}

 		/// <summary>
 		/// Properties context menu clicked.
 		/// </summary>
 		public void OnPropertiesClick(object sender, EventArgs e)
 		{
			if(pDialog != null && ! pDialog.IsDisposed)
				// Already open
				return;

			// Display the dialog
			pDialog = new propertiesDialog(this);
			pDialog.Show();

 		}

		/// <summary>
		/// Properties Dialog
		/// </summary>
		public class propertiesDialog : System.Windows.Forms.Form
		{
			private System.Windows.Forms.Label lblFactors;
			private System.Windows.Forms.ComboBox cboFactors;
			private System.Windows.Forms.Button btnOK;
			private System.Windows.Forms.Button btnCancel;
			private FogLayer layer;

			public propertiesDialog( FogLayer layer )
			{
				InitializeComponent();
				//this.Icon = WorldWind.PluginEngine.Plugin.Icon;
				this.layer = layer;
				// Init factors list presets
				// TODO: use fogPreset objects
				string[] presets = {"Very light", "Light", "Medium", "Thick", "London"};
				cboFactors.Items.AddRange(presets);
				if(layer.nearFactor >= 1.0f) cboFactors.SelectedIndex = 4-(int)layer.nearFactor; 
				if(layer.nearFactor <  1.0f) cboFactors.SelectedIndex = 4; 
				// select current factors
				//int i = cboFactors.FindString(layer.textureFileName);
				//if(i != -1) cboFactors.SelectedIndex = i;
			}

			#region Windows Form Designer generated code
			/// <summary>
			/// Required method for Designer support - do not modify
			/// the contents of this method with the code editor.
			/// </summary>
			private void InitializeComponent()
			{
				this.btnCancel = new System.Windows.Forms.Button();
				this.btnOK = new System.Windows.Forms.Button();
				this.lblFactors = new System.Windows.Forms.Label();
				this.cboFactors = new System.Windows.Forms.ComboBox();
				this.SuspendLayout();
				// 
				// btnCancel
				// 
				this.btnCancel.DialogResult = System.Windows.Forms.DialogResult.Cancel;
				this.btnCancel.Location = new System.Drawing.Point(311, 59);
				this.btnCancel.Name = "btnCancel";
				this.btnCancel.TabIndex = 0;
				this.btnCancel.Text = "Cancel";
				this.btnCancel.Click += new System.EventHandler(this.btnCancel_Click);
				// 
				// btnOK
				// 
				this.btnOK.Location = new System.Drawing.Point(224, 59);
				this.btnOK.Name = "btnOK";
				this.btnOK.TabIndex = 1;
				this.btnOK.Text = "OK";
				this.btnOK.Click += new System.EventHandler(this.btnOK_Click);
				// 
				// lblFactors
				// 
				this.lblFactors.AutoSize = true;
				this.lblFactors.Location = new System.Drawing.Point(16, 28);
				this.lblFactors.Name = "lblFactors";
				this.lblFactors.Size = new System.Drawing.Size(82, 16);
				this.lblFactors.TabIndex = 2;
				this.lblFactors.Text = "Fog intensity:";
				// 
				// cboFactors
				// 
				this.cboFactors.Location = new System.Drawing.Point(96, 25);
				this.cboFactors.Name = "cboFactors";
				this.cboFactors.Size = new System.Drawing.Size(296, 21);
				this.cboFactors.TabIndex = 3;
				//this.cboFactors.Text = "Select fog";
				this.cboFactors.DropDownStyle = ComboBoxStyle.DropDownList;
				this.cboFactors.MaxDropDownItems = 10;
				// 
				// frmFavorites
				// 
				this.AcceptButton = this.btnOK;
				this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
				this.CancelButton = this.btnCancel;
				this.ClientSize = new System.Drawing.Size(406, 94);
				this.ControlBox = false;
				this.Controls.Add(this.cboFactors);
				this.Controls.Add(this.lblFactors);
				this.Controls.Add(this.btnOK);
				this.Controls.Add(this.btnCancel);
				this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.FixedToolWindow;
				this.MaximizeBox = false;
				this.MinimizeBox = false;
				this.Name = "pDialog";
				this.ShowInTaskbar = false;
				this.StartPosition = System.Windows.Forms.FormStartPosition.CenterScreen;
				//this.StartPosition = System.Windows.Forms.FormStartPosition.Manual;
				//this.Location = new System.Drawing.Point(layer.drawArgs.CurrentMousePosition.X + 10, layer.drawArgs.CurrentMousePosition.Y - 10);
				this.Text = "Fog properties";
				this.TopMost = true;
				this.ResumeLayout(false);

			}
			#endregion

			private void btnOK_Click(object sender, System.EventArgs e)
			{
				if(cboFactors.SelectedItem != null) 
				{
					// Set factors
					// TODO: use fogPreset objects
					switch (cboFactors.SelectedItem.ToString())
					{
						case "Very light": 	layer.nearFactor = 4.0f; layer.farFactor = 40.0f; break;
						case "Light": 		layer.nearFactor = 3.0f; layer.farFactor = 25.0f; break;
						case "Medium": 		layer.nearFactor = 2.0f; layer.farFactor = 15.0f; break;
						case "Thick": 		layer.nearFactor = 1.0f; layer.farFactor = 10.0f; break;
						case "London": 		layer.nearFactor = 0.5f; layer.farFactor = 7.0f; break;
					}
  					layer.SaveSettings();
				}
				// Close this form
				this.Close();
			}

			private void btnCancel_Click(object sender, System.EventArgs e)
			{

				// Close this form
				this.Close();
			}



		}

		#endregion
	}
}
